/**
 * Authors: http://d.hatena.ne.jp/ultraist/
 */
module pes.variant;

/**
 * Variant Type
 *
 *
 */

private import std.stdarg;
private import std.stdio;
private import std.math2;
private import std.conv;
private import std.ctype;

private import pes.exception;
private import pes.conv;

/* static NULL Variant */
public static VNull VNULL;
static this()
{
	VNULL = new VNull;
}

/**
 * Variant
 *
 *
 *
 *
 *
 **/
abstract class Variant
{
protected:
	static Variant make(TypeInfo type_info, inout va_list argptr)
	{
		if (type_info is typeid(Variant)) {
			return va_arg!(Variant)(argptr);
		} else if (type_info is typeid(long)) {
			return new VLong(va_arg!(long)(argptr));
		} else if (type_info is typeid(real)) {
			return  new VReal(va_arg!(real)(argptr));
		} else if (type_info is typeid(char[])) {
			return  new VString(va_arg!(char[])(argptr));
		} else if (type_info is typeid(bool)) {
			return new VBool(va_arg!(bool)(argptr));
		}
		
		return null;
	}
	
	static Variant make(...)
	{
		if (_arguments.length == 1) {
			return make(_arguments[0], _argptr);
		}
		
		return VArray.make(_arguments, _argptr);
	}
	
	static Variant make(char [] value)
	{
		real	real_value;
		long	long_value;
		bool	is_long;
		
		/* test null */
		if (value.length == 0) {
			return new VNull;
		}
		
		/* test Long */
		is_long = true;
		try {
			long_value = stringToLong(value);
		} catch (Object) {
			is_long = false;
		}
		if (is_long) {
			return new VLong(long_value);
		}
		
		/* test Real */
		real_value = std.math2.atof(value);
		if (! std.math.isnan(real_value)) {
			return new VReal(real_value);
		}
		
		return new VString(value); // VString
	}
	
	/*--------------------------------
	 + exception
	 +--------------------------------*/
	static void notImplementedException(char [] name, char [] function_name)
	{
		throw new NotImplemented(name, function_name);
	}
	
	void notImplementedException(char [] function_name)
	{
		notImplementedException(this.classinfo.name, function_name);
	}
	
	void convertException(char [] message)
	{
		throw new ConvError(message);
	}
	
public:
	enum NumericType
	{
		Nan,
		Long,
		Real,
		Auto
	}
	
	/*--------------------------------
	 + tester
	 +--------------------------------*/
	NumericType numericType()
	{
		return NumericType.Nan;
	}
	
	bool isNumeric()
	{
		return numericType != NumericType.Nan;
	}
	
	abstract TypeInfo type();
	
	/*--------------------------------
	 + converter
	 +--------------------------------*/
	
	Object toObject()
	{
		notImplementedException("toObject");
		return null;
	}
	
	long toLong()
	{
		notImplementedException("(toLong|toReal)");
		return 0;
	}
	
	real toReal()
	{
		real real_value = cast(real)toLong;
		
		if (std.math.isnan(real_value)) {
			notImplementedException("(toLong|toReal)");
		}
		
		return real_value;
	}
	
	char [] toString()
	{
		notImplementedException("toString");
		return null;
	}
	
	bool toBool()
	{
		notImplementedException("toBool");
		return false;
	}
	
	bool toBool(Variant true_value, Variant false_value)
	in {
		assert(true_value == null && false_value == null);
	}
	body {
		bool 	bool_value;
		
		if (true_value == null) {
			if (compareString(false_value) == 0) {
				bool_value = false;
			} else {
				bool_value = true;
			}
		} else if (false_value == null) {
			if (compareString(true_value) == 0) {
				bool_value = true;
			} else {
				bool_value = false;
			}
		} else {
			if (compareString(true_value) == 0) {
				bool_value = true;
			} else if (compareString(false_value) == 0) {
				bool_value = false;
			} else {
				convertException(
					std.string.format(
						"`%s' was not able to be converted to type bool(true = %s, false = %s)",
						toString, true_value, false_value
					));
			}
		}
		
		return bool_value;
	}
	
	Variant opIndex(Variant n)
	{
		notImplementedException("opIndex");
		return null;
	}
	
	Variant opIndexAssign(Variant value, Variant n)
	{
		notImplementedException("opIndexAssign");
		return null;
	}
	
	Variant opCall(Variant argv)
	{
		notImplementedException("opCall");
		return null;
	}
	
	/*--------------------------------
	 + numeric operation
	 +--------------------------------*/
	
	
	/*	Binary Operation	*/
	
	Variant assign(Variant rhs)
	{
		notImplementedException("assign");
		return null;
	}
	
	Variant opAdd(Variant rhs)
	{
		if (numericType == NumericType.Real
			|| rhs.numericType == NumericType.Real) 
		{
			return new VReal(toReal + rhs.toReal);
		}
		
		return new VLong(toLong + rhs.toLong);
	}
	
	Variant opSub(Variant rhs)
	{
		if (numericType == NumericType.Real
			|| rhs.numericType == NumericType.Real)
		{
			return new VReal(toReal - rhs.toReal);
		}
		
		return new VLong(toLong - rhs.toLong);
	}
	
	Variant opMul(Variant rhs)
	{
		if (numericType == NumericType.Real
			|| rhs.numericType == NumericType.Real)
		{
			return new VReal(toReal * rhs.toReal);
		}
		
		return new VLong(toLong * rhs.toLong);
	}
	
	Variant opDiv(Variant rhs)
	{
		if (numericType == NumericType.Real
			|| rhs.numericType == NumericType.Real)
		{
			return new VReal(toReal / rhs.toReal);
		}
		
		return new VLong(toLong / rhs.toLong);
	}
	
	Variant opMod(Variant rhs)
	{
		if (numericType == NumericType.Real
			|| rhs.numericType == NumericType.Real)
		{
			return new VReal(toReal % rhs.toReal);
		}
		
		return new VLong(toLong % rhs.toLong);
	}
	
	Variant opAnd(Variant rhs)
	{
		return new VLong(toLong & rhs.toLong);
	}
	
	Variant opOr(Variant rhs)
	{
		return new VLong(toLong | rhs.toLong);
	}
	
	Variant opXor(Variant rhs)
	{
		return new VLong(toLong ^ rhs.toLong);
	}
	
	Variant opShl(Variant rhs)
	{
		return new VLong(toLong << rhs.toLong);
	}
	
	Variant opShr(Variant rhs)
	{
		return new VLong(toLong >> rhs.toLong);
	}
	
	Variant opUShr(Variant rhs)
	{
		return new VLong(toLong >>> rhs.toLong);
	}
	
	Variant opCom()
	{
		return new VLong(~toLong);
	}
	
	
	/*	Unary Operation	*/
	
	Variant opPostInc()
	{
		return assign(new VLong(toLong + 1));
	}
	
	Variant opPostDec()
	{
		return assign(new VLong(toLong - 1));
	}
	
	Variant opNeg()
	{
		if (numericType == NumericType.Real) {
			return new VReal(-toReal);
		}
		
		return new VLong(-toLong);
	}
	
	Variant opPos()
	{
		if (numericType == NumericType.Real) {
			return new VReal(+toReal);
		}
		
		return new VLong(+toLong);
	}
	
	/* Compare operation */
	
//	int opEauals(Variant rhs)
//	{
//		return opCmp(rhs) == 0;
//	}
	
	int equals(Variant rhs)
	{
		return opCmp(rhs) == 0;
	}
	
	int opCmp(Variant rhs)
	{
		if (numericType == NumericType.Real
			|| rhs.numericType == NumericType.Real)
		{
			real diff_real = toReal - rhs.toReal;
			
			if (diff_real > 0.0) {
				return 1;
			} else if (diff_real < 0.0) {
				return -1;
			}
			
			return 0;
		}
		
		long diff_long = toLong - rhs.toLong;
		
		if (diff_long > 0) {
			return 1;
		} else if (diff_long < 0) {
			return -1;
		}
		
		return 0;
	}
	
	/*--------------------------------
	 + string operation
	 +--------------------------------*/
	int compareString(Variant rhs)
	{
		return std.string.cmp(toString, rhs.toString);
	}
	
	Variant appendString(Variant rhs)
	{
		return new VString(toString ~ rhs.toString);
	}
}


/**
 * VNull: Variant
 *
 *
 *
 *
 *
 **/
class VNull: Variant
{
	static this() {}
	
	long toLong()
	{
		return 0L;
	}
	
	real toReal()
	{
		return 0.0L;
	}
	
	char [] toString()
	{
		return "";
	}
	
	bool toBool()
	{
		return false;
	}
	
	TypeInfo type()
	{
		return typeid(VNull);
	}
}

/**
 *
 * VObject: Variant
 *
 *
 *
 *
 **/
class VObject(T: Object): Variant
{
protected:
	T	m_value;
	
public:
	this(T value)
	{
		m_value = value;
	}
	
	Variant assign(Variant rhs)
	{
		T test_object = cast(T)rhs.toObject;
		if (test_object is null) {
			convertException(
				rhs.toObject.classinfo.toString ~ ".to" ~ m_value.classinfo.toString
			);
		}
		m_value = test_object;
		
		return this;
	}
	
	Variant assign(T rhs)
	{
		m_value = rhs;
		return this;
	}
	
	Object toObject()
	{
		return cast(Object )m_value;
	}
	
	long toLong()
	{
		static if (is(T : long)) {
			return cast(long )m_value;
		} else {
			return super.toLong;
		}
	}
	
	real toReal()
	{
		static if (is(T: real)) {
			return cast(real )m_value;
		} else {
			return super.toReal;
		}
	}
	
	char [] toString()
	{
		return (cast(Object)(m_value)).toString;
	}
	
	bool toBool()
	{
		static if (is(T: bool)) {
			return cast(bool)m_value;
		} else {
			return super.toBool;
		}
	}
	
	TypeInfo type()
	{
		return typeid(T);
	}
}

class VLong: Variant
{
protected:
	long	m_value;

public:
	this(long value)
	{
		m_value = value;
	}
	
	this(VLong value)
	{
		m_value = value.m_value;
	}
	
	Variant assign(Variant rhs)
	{
		m_value = rhs.toLong;
		return this;
	}
	
	Variant assign(long rhs)
	{
		m_value = rhs;
		return this;
	}
	
	NumericType numericType()
	{
		return NumericType.Long;
	}
	
	long toLong()
	{
		return m_value;
	}
	
	real toReal()
	{
		return cast(real )m_value;
	}
	
	char [] toString()
	{
		return std.string.toString(m_value);
	}
	
	bool toBool()
	{
		return m_value ? true: false;
	}
	
	TypeInfo type()
	{
		return typeid(VLong);
	}
}


/**
 * Variant.VReal
 *
 *
 *
 *
 *
 **/
class VReal: Variant
{
protected:
	real	m_value;

public:
	this(real value)
	{
		m_value = value;
	}
	
	this(VReal value)
	{
		m_value = value.m_value;
	}
	
	Variant assign(Variant rhs)
	{
		m_value = rhs.toReal;
		return this;
	}
	
	Variant assign(real rhs)
	{
		m_value = rhs;
		return this;
	}
	
	NumericType numericType()
	{
		return NumericType.Real;
	}
	
	long toLong()
	{
		return cast(long )m_value;
	}
	
	real toReal()
	{
		return m_value;
	}
	
	char [] toString()
	{
		return std.string.toString(m_value);
	}
	
	bool toBool()
	{
		return m_value ? true: false;
	}
	
	TypeInfo type()
	{
		return typeid(VReal);
	}
}


/**
 * VString: Variant
 *
 *
 *
 *
 *
 **/
class VString: Variant
{
protected:
	char [] m_value;

public:
	this(char [] value)
	{
		m_value = value.dup;
	}
	
	this(VString value)
	{
		m_value = value.m_value.dup;
	}
	
	Variant assign(Variant rhs)
	{
		m_value = rhs.toString;
		return this;
	}
	
	Variant assign(char [] rhs)
	{
		m_value = rhs.dup;
		return this;
	}
	
	NumericType numericType()
	{
		real	real_value;
		bool	is_long;
		long	long_value;
		
		/* test VReal */
		real_value = std.math2.atof(m_value);
		if (! std.math.isnan(real_value)) {
			return NumericType.Real;
		}
		
		/* test Long */
		is_long = true;
		try {
			long_value = stringToLong(m_value);
		} catch (Object) {
			is_long = false;
		}
		if (is_long) {
			return NumericType.Long;
		}
		
		return NumericType.Nan;
	}
	
	long toLong()
	{
		long long_value;
		
		try {
			long_value = stringToLong(m_value);
		} catch (Object) {
			convertException(toString ~ ".toLong");
		}
		
		return long_value;
	}
	
	real toReal()
	{
		real value = std.math2.atof(m_value);
		
		if (std.math.isnan(value)) {
			convertException(toString ~ ".toReal");
		}
		
		return value;
	}
	
	bool toBool()
	{
		return m_value.length ? true: false;
	}
	
	char [] toString()
	{
		return m_value;
	}
	
	Variant opIndex(Variant i)
	{
		long index = i.toLong;
		
		if (index > 0 && m_value.length > index) {
			return new VString(m_value[index .. index + 1]);
		}
		
		return new VNull; /// throw new ArrayBoundsError;
	}
	
	Variant opIndexAssign(Variant value, Variant i)
	{
		long index = i.toLong;
		
		if (index > 0 && m_value.length > index) {
			char[] charv = value.toString;
			
			if (charv.length == 0) {
				/// access valid
			}
			m_value[index] = charv[0];
		}
		
		return this;
	}
	
	TypeInfo type()
	{
		return typeid(VString);
	}
}


/**
 * VBool: Variant
 *
 *
 *
 *
 *
 **/
class VBool: Variant
{
protected:
	bool m_value;

public:
	this(bool value)
	{
		m_value = value;
	}
	
	this(VBool value)
	{
		m_value = value.m_value;
	}
	
	Variant assign(Variant rhs)
	{
		m_value = rhs.toBool;
		return this;
	}
	
	Variant assign(bool rhs)
	{
		m_value = rhs;
		return this;
	}
	
	long toLong()
	{
		return m_value == true ? 1: 0;
	}
	
	real toReal()
	{
		return m_value == true ? 1.0: 0.0;
	}
	
	bool toBool()
	{
		return m_value;
	}
	
	char [] toString()
	{
		return (m_value ? "true":"false").dup;
	}
	
	TypeInfo type()
	{
		return typeid(VBool);
	}
}


/**
 * VArray: Variant
 *
 *
 *
 *
 *
 **/
class VArray: Variant
{
protected:
	Variant[] m_array;
	
public:
	this(uint n)
	{
		m_array = new Variant[n];
		for (int i = 0; i < n; ++i) {
			m_array[i] = VNULL;
		}
	}
	
	this(Variant [] array)
	{
		m_array = array.dup;
	}
	
	this(VArray array)
	{
		m_array = array.m_array.dup;
	}
	
	NumericType numericType()
	{
		return NumericType.Long; // length
	}
	
	static VArray make(...)
	{
		return VArray.make(_arguments, _argptr);
	}
	
	static VArray make(TypeInfo [] type_info, va_list argptr)
	{
		Variant [] 	array;
		Variant 	value;
		
		for (uint i = 0; i < type_info.length; ++i) {
			value = Variant.make(type_info[i], argptr);
			if (value is null) {
				notImplementedException(type_info[i].toString, 
											"convert(" ~ type_info[i].toString ~ ")");
			}
			array ~= value;
		}
		
		return new VArray(array);
	}
	
	Variant assign(Variant rhs)
	{
		VArray test_array = cast(VArray)rhs;
		
		if (test_array is null) {
			convertException(rhs.type.toString ~ ".toVArray");
		}
		
		m_array = test_array.m_array.dup;
		
		return this;
	}
	
	Variant assign(VArray rhs)
	{
		m_array = rhs.m_array.dup;
		return this;
	}
	
	Variant opIndex(int i)
	{
		for (long j = m_array.length; j <= i; ++j) {
			m_array ~= VNULL;
		}
		
		return m_array[i];
	}
	
	Variant opIndex(Variant i)
	{
		return opIndex(cast(int)i.toLong);
	}
	
	Variant opIndexAssign(Variant value, int i)
	{
		for (int j = m_array.length; j <= i; ++j) {
			m_array ~= VNULL;
		}
		
		return (m_array[i] = value);
	}
	
	Variant opIndexAssign(Variant value, Variant i)
	{
		return opIndexAssign(value, cast(int)i.toLong);
	}
	
	long toLong()
	{
		return m_array.length;
	}
	
	char [] toString()
	{
		char [] join_elm;
		
		foreach (Variant v; m_array) {
			join_elm ~= v.toString;
		}
		
		return join_elm;
	}
	
	TypeInfo type()
	{
		return typeid(VArray);
	}
	
	//
	VArray append(VArray array)
	{
		m_array ~= array.m_array.dup;
		return this;
	}
	
	Variant pop()
	{
		Variant popv = m_array[length];
		m_array.length = m_array.length - 1;
		
		return popv;
	}
	
	Variant push(Variant pushv)
	{
		m_array ~= pushv;
		return pushv;
	}
	
	VArray dup()
	{
		return new VArray(m_array);
	}
	
	Variant[] dup_internal()
	{
		return m_array.dup;
	}
}

/**
 * DEBUG pes_variant
 * 
 **/
debug(pes_variant)
{
	private import std.stream;
	int main()
	{
		{
			File create_test_file = new File("variant_debug_out.txt", FileMode.Out);
			create_test_file.writeLine("HELLO VARIANT");
			create_test_file.close();
		}
		
		/* basic type */
		Variant longv = new VLong(5);
		Variant realv = new VReal(6.4);
		Variant stringv = new VString("8.1234");
		
		writef("expr: %s + %s + %s = %s\n", longv, realv, stringv, longv + realv + stringv);
		
		/* VObject */
		alias VObject!(Stream) VStream;
		Variant stream = new VStream(new File("variant_debug_out.txt", FileMode.In));
		
		/* VArray */
		Variant array = VArray.make(1L,  0.14L, true, "1.8piyo", stream);
		array[longv] = new VString("hoge");
		array[realv] = new VLong(2);
		
		writef("\n-array[%s]\n", array);
		Variant sumv = new VLong(0);
		for (Variant j = new VLong(0); (new VBool(j < array)).toBool; j++) {// Bool lol
			writef("%d -> %s (%s)\n", j, array[j], array[j].type.toString);
			if (array[j].isNumeric) {
				// numeric
				sumv = sumv + array[j];	// ((Variant + Variant)	-> Variant
			}
		}
		
		writef("\nsum of numeric value -> %s(%s)\n",
					sumv.type.toString, sumv);
		
		// auto convertion
		writef("\n-- auto convertion\n");
		VArray convtest1 = VArray.make(
									1.175494351e-38L,
									1.175494351L,
									1.0000000L,
									1_2_3_4_5_6L,
									+123456L,
									Variant.make("1.175494351e-38"),
									Variant.make("1.175494351"),
									Variant.make("1.0000000"),
									Variant.make("1_2_3_4_5_6"),
									Variant.make("+123456")
							);
		VArray convtest2 = VArray.make(
									-123456L,
									-0x1234L,
									+01234L,
									"hoge",
									Variant.make("-123456"),
									Variant.make("-0x1234"),		// hex
									Variant.make("+01234"),			// oct
									Variant.make("hoge")
								);
		Variant convtest = convtest1.append(convtest2).dup;
		
		for (Variant j = new VLong(0); j < convtest; j++) {
			writef("%s to %s\n", convtest[j], convtest[j].type.toString);
		}
		
		return 0;
	}
}

